# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# © 2023 ANSYS, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing types for the Layer Service."""


from typing import Optional, Union

from ansys.api.sherlock.v0 import SherlockLayerService_pb2
from pydantic import BaseModel, ValidationInfo, field_validator, model_validator
from typing_extensions import Self

from ansys.sherlock.core.types.common_types import basic_str_validator, optional_str_validator


class PolygonalShape(BaseModel):
    """Contains the properties for a polygonal shape."""

    points: list[tuple[float, float]]
    """Points (length two tuples of the form (x, y)) : list[tuple[float, float]]"""
    rotation: float
    """Rotation (in degrees) : float"""

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.PolygonalShape:
        grpc_polygonal_shape = SherlockLayerService_pb2.PolygonalShape()
        for x, y in self.points:
            p = grpc_polygonal_shape.points.add()
            p.x = x
            p.y = y
        grpc_polygonal_shape.rotation = self.rotation
        return grpc_polygonal_shape


class RectangularShape(BaseModel):
    """Contains the properties for a rectangular shape."""

    length: float
    """Length : float"""
    width: float
    """Width : float"""
    center_x: float
    """X coordinate of center : float"""
    center_y: float
    """Y coordinate of center : float"""
    rotation: float
    """Rotation (in degrees) : float"""

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.RectangularShape:
        grpc_rectangular_shape = SherlockLayerService_pb2.RectangularShape()
        grpc_rectangular_shape.length = self.length
        grpc_rectangular_shape.width = self.width
        grpc_rectangular_shape.centerX = self.center_x
        grpc_rectangular_shape.centerY = self.center_y
        grpc_rectangular_shape.rotation = self.rotation
        return grpc_rectangular_shape


class SlotShape(BaseModel):
    """Contains the properties for a slot shape."""

    length: float
    """Length : float"""
    width: float
    """Width : float"""
    node_count: int
    """Node count : int"""
    center_x: float
    """X coordinate of center : float"""
    center_y: float
    """Y coordinate of center : float"""
    rotation: float
    """Rotation (in degrees) : float"""

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.SlotShape:
        grpc_slot_shape = SherlockLayerService_pb2.SlotShape()

        grpc_slot_shape.length = self.length
        grpc_slot_shape.width = self.width
        grpc_slot_shape.nodeCount = self.node_count
        grpc_slot_shape.centerX = self.center_x
        grpc_slot_shape.centerY = self.center_y
        grpc_slot_shape.rotation = self.rotation

        return grpc_slot_shape


class CircularShape(BaseModel):
    """Contains the properties for a circular shape."""

    diameter: float
    """Diameter : float"""
    node_count: int
    """Node count : int"""
    center_x: float
    """X coordinate of center : float"""
    center_y: float
    """Y coordinate of center : float"""
    rotation: float
    """Rotation (in degrees) : float"""

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.CircularShape:
        grpc_circular_shape = SherlockLayerService_pb2.CircularShape()
        grpc_circular_shape.diameter = self.diameter
        grpc_circular_shape.nodeCount = self.node_count
        grpc_circular_shape.centerX = self.center_x
        grpc_circular_shape.centerY = self.center_y
        grpc_circular_shape.rotation = self.rotation
        return grpc_circular_shape


class PCBShape(BaseModel):
    """Contains the properties for a PCB shape."""

    @staticmethod
    def _convert_to_grpc() -> SherlockLayerService_pb2.PCBShape:
        return SherlockLayerService_pb2.PCBShape()


class PottingRegion(BaseModel):
    """Contains the properties of a Potting Region add or update request."""

    cca_name: str
    """The name of the CCA."""
    potting_id: str
    """The potting ID."""
    potting_side: str
    """The potting side, options are "TOP", "BOT", or "BOTTOM"."""
    potting_material: str
    """The potting material."""
    potting_units: str
    """The units to use for the potting region."""
    potting_thickness: float
    """The potting thickness."""
    potting_standoff: float
    """The potting standoff."""
    shape: Union[CircularShape, PCBShape, PolygonalShape, RectangularShape, SlotShape]
    """The shape of the potting region."""

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.PottingRegion:
        grpc_potting_region_data = SherlockLayerService_pb2.PottingRegion()

        grpc_potting_region_data.ccaName = self.cca_name
        grpc_potting_region_data.pottingID = self.potting_id
        grpc_potting_region_data.pottingSide = self.potting_side
        grpc_potting_region_data.pottingMaterial = self.potting_material
        grpc_potting_region_data.pottingUnits = self.potting_units
        grpc_potting_region_data.pottingThickness = self.potting_thickness
        grpc_potting_region_data.pottingStandoff = self.potting_standoff

        if isinstance(self.shape, CircularShape):
            grpc_potting_region_data.circularShape.CopyFrom(self.shape._convert_to_grpc())
        elif isinstance(self.shape, PCBShape):
            grpc_potting_region_data.pCBShape.CopyFrom(self.shape._convert_to_grpc())
        elif isinstance(self.shape, PolygonalShape):
            grpc_potting_region_data.polygonalShape.CopyFrom(self.shape._convert_to_grpc())
        elif isinstance(self.shape, RectangularShape):
            grpc_potting_region_data.rectangularShape.CopyFrom(self.shape._convert_to_grpc())
        elif isinstance(self.shape, SlotShape):
            grpc_potting_region_data.slotShape.CopyFrom(self.shape._convert_to_grpc())
        else:
            raise ValueError("Unsupported shape given '" + type(self.shape).__name__ + "'")

        return grpc_potting_region_data

    @field_validator("cca_name", "potting_id", "potting_side", "potting_material", "potting_units")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)


class PottingRegionUpdateData(BaseModel):
    """Contains the properties of a potting region update request."""

    potting_region_id_to_update: str
    """ID of the potting region to update."""
    potting_region: PottingRegion
    """Potting region data used to update the potting region."""

    def _convert_to_grpc(
        self,
    ) -> SherlockLayerService_pb2.UpdatePottingRegionRequest.PottingRegionUpdateData:

        grpc_potting_update_data = (
            SherlockLayerService_pb2.UpdatePottingRegionRequest.PottingRegionUpdateData()
        )

        grpc_potting_update_data.pottingRegionIDToUpdate = self.potting_region_id_to_update
        grpc_potting_update_data.pottingRegion.CopyFrom(self.potting_region._convert_to_grpc())
        return grpc_potting_update_data

    @field_validator("potting_region_id_to_update")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)


class UpdatePottingRegionRequest(BaseModel):
    """Contains the properties of a potting region update per project."""

    project: str
    """Name of the Sherlock project."""
    update_potting_regions: list[PottingRegionUpdateData]
    """List of potting region data to update."""

    @field_validator("project")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(
        self,
    ) -> SherlockLayerService_pb2.UpdatePottingRegionRequest:
        request = SherlockLayerService_pb2.UpdatePottingRegionRequest()
        request.project = self.project
        for update_potting_region in self.update_potting_regions:
            request.updatePottingRegions.append(update_potting_region._convert_to_grpc())
        return request


class PottingRegionCopyData(BaseModel):
    """Data identifying which potting regions to copy."""

    cca_name: str
    """Name of the cca."""
    potting_id: str
    """ID to assign to the new potting region."""
    copy_potting_id: str
    """ID of the potting region to copy."""
    center_x: float
    """X coordinate for the center of the new potting region."""
    center_y: float
    """Y coordinate for the center of the new potting region."""

    @field_validator("cca_name", "potting_id", "copy_potting_id")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    @model_validator(mode="after")
    def validate_ids(self) -> Self:
        """Validate that the potting IDs are not the same between the original and copy."""
        if self.potting_id == self.copy_potting_id:
            raise ValueError("potting_id must be different than copy_potting_id")
        return self

    def _convert_to_grpc(
        self,
    ) -> SherlockLayerService_pb2.CopyPottingRegionRequest.PottingRegionCopyData:

        grpc_potting_copy_data = (
            SherlockLayerService_pb2.CopyPottingRegionRequest.PottingRegionCopyData()
        )

        grpc_potting_copy_data.ccaName = self.cca_name
        grpc_potting_copy_data.pottingID = self.potting_id
        grpc_potting_copy_data.copyPottingID = self.copy_potting_id
        grpc_potting_copy_data.centerX = self.center_x
        grpc_potting_copy_data.centerY = self.center_y

        return grpc_potting_copy_data


class CopyPottingRegionRequest(BaseModel):
    """Request to delete 1 or more potting regions."""

    project: str
    """Name of the project containing the potting region(s) to be copied."""
    potting_region_copy_data: list[PottingRegionCopyData]
    """Data identifying which potting regions to copy and what potting regions to copy from."""

    @field_validator("project")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(
        self,
    ) -> SherlockLayerService_pb2.CopyPottingRegionRequest:

        grpc_potting_copy_request = SherlockLayerService_pb2.CopyPottingRegionRequest()
        grpc_potting_copy_request.project = self.project
        for copy_data in self.potting_region_copy_data:
            grpc_potting_copy_request.pottingRegionCopyData.append(copy_data._convert_to_grpc())

        return grpc_potting_copy_request


class PottingRegionDeleteData(BaseModel):
    """Data specifying potting regions to delete."""

    cca_name: str
    """Name of the CCA containing the potting region(s) to delete."""
    potting_id: str
    """Id of the potting region(s) to delete."""

    @field_validator("cca_name", "potting_id")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(
        self,
    ) -> SherlockLayerService_pb2.DeletePottingRegionRequest.PottingRegionDeleteData:
        delete_data = SherlockLayerService_pb2.DeletePottingRegionRequest.PottingRegionDeleteData()

        delete_data.ccaName = self.cca_name
        delete_data.pottingID = self.potting_id
        return delete_data


class DeletePottingRegionRequest(BaseModel):
    """Request to delete 1 or more potting regions."""

    project: str
    """Name of the project containing the potting regions to delete."""
    potting_region_delete_data: list[PottingRegionDeleteData]
    """Data identifying which potting regions should be deleted."""

    @field_validator("project")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.DeletePottingRegionRequest:
        request = SherlockLayerService_pb2.DeletePottingRegionRequest()
        request.project = self.project
        for delete_data in self.potting_region_delete_data:
            request.pottingRegionDeleteData.append(delete_data._convert_to_grpc())
        return request


class GetTestPointPropertiesRequest(BaseModel):
    """Return the properties for each test point given a comma-separated list of test point ids."""

    project: str
    """Name of the project."""
    cca_name: str
    """Name of the CCA containing the test point properties to return."""
    test_point_ids: Optional[str] = None
    """Optional Param: Comma-separated list of test point ids representing one or more test points.
        If this parameter is not included, then the entire list of test points for a given CCA will
        have their properties returned.
    """

    @field_validator("project", "cca_name", "test_point_ids")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    @field_validator("test_point_ids")
    @classmethod
    def optional_str_validation(cls, value: Optional[str], info):
        """Allow the test_point_ids to not be set, i.e., None."""
        return optional_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.GetTestPointPropertiesRequest:
        request = SherlockLayerService_pb2.GetTestPointPropertiesRequest()
        request.project = self.project
        request.ccaName = self.cca_name
        if self.test_point_ids is not None:
            request.testPointIDs = self.test_point_ids
        return request


class GetICTFixturesPropertiesRequest(BaseModel):
    """Return the properties for each ICT fixture given a comma-separated list of fixture ids."""

    project: str
    """Name of the project."""
    cca_name: str
    """Name of the CCA containing the ict fixture properties to return."""
    ict_fixtures_ids: Optional[str] = None
    """Optional Param: Comma-separated list of ict fixture ids representing one or more ict
        fixtures.   If this parameter is not included, then the entire list of ict fixtures
        for a given CCA will have their properties returned.
    """

    @field_validator("project", "cca_name", "ict_fixtures_ids")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    @field_validator("ict_fixtures_ids")
    @classmethod
    def optional_str_validation(cls, value: Optional[str], info):
        """Allow the ict_fixtures_ids to not be set, i.e., None."""
        return optional_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.GetICTFixturesPropertiesRequest:
        request = SherlockLayerService_pb2.GetICTFixturesPropertiesRequest()
        request.project = self.project
        request.ccaName = self.cca_name
        if self.ict_fixtures_ids is not None:
            request.ICTFixtureIDs = self.ict_fixtures_ids
        return request


class TestPointProperties(BaseModel):
    """Contains the properties of a test point."""

    __test__ = False  # This line is to notify pytest that this is not a test class.

    id: str
    """ID"""
    side: str
    """Side"""
    units: str
    """Units"""
    center_x: float
    """Center x-value"""
    center_y: float
    """Center y-value"""
    radius: float
    """Radius"""
    load_type: SherlockLayerService_pb2.TestPointProperties.LoadType.ValueType
    """Load type"""
    load_value: float
    """Load value"""
    load_units: str
    """Load units"""

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.TestPointProperties:
        grpc_test_point_data = SherlockLayerService_pb2.TestPointProperties()

        grpc_test_point_data.ID = self.id
        grpc_test_point_data.side = self.side
        grpc_test_point_data.units = self.units
        grpc_test_point_data.centerX = self.center_x
        grpc_test_point_data.centerY = self.center_y
        grpc_test_point_data.radius = self.radius
        grpc_test_point_data.loadType = self.load_type
        grpc_test_point_data.loadValue = self.load_value
        grpc_test_point_data.loadUnits = self.load_units

        return grpc_test_point_data

    @field_validator("side", "units", "load_type")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)


class UpdateTestPointsRequest(BaseModel):
    """Contains the properties of a test points update per project."""

    project: str
    """Name of the Sherlock project."""
    cca_name: str
    """Name of the Sherlock CCA."""
    update_test_points: list[TestPointProperties]
    """List of test points with their properties to update"""

    @field_validator("project", "cca_name")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.UpdateTestPointsRequest:
        request = SherlockLayerService_pb2.UpdateTestPointsRequest()
        request.project = self.project
        request.ccaName = self.cca_name
        if self.update_test_points is not None:
            for update_test_point in self.update_test_points:
                request.testPointProperties.append(update_test_point._convert_to_grpc())
        return request


class ICTFixtureProperties(BaseModel):
    """Contains the properties of an ict fixture."""

    id: str
    """ID"""
    type: str
    """Type"""
    units: str
    """Units"""
    side: str
    """Side"""
    height: str
    """Height"""
    material: str
    """Material"""
    state: str
    """State"""
    shape: str
    """Shape type"""
    x: str
    """Center X"""
    y: str
    """Center Y"""
    length: str
    """Length"""
    width: str
    """Width"""
    diameter: str
    """Diameter"""
    nodes: str
    """Number of nodes"""
    rotation: str
    """Degrees of rotation"""
    polygon: str
    """Coordinates of points"""
    boundary: str
    """Boundary point(s)"""
    constraints: str
    """FEA constraints"""
    chassis_material: str
    """Chassis material"""

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.ICTFixtureProperties:
        grpc_fixture_data = SherlockLayerService_pb2.ICTFixtureProperties()

        grpc_fixture_data.ID = self.id
        grpc_fixture_data.type = self.type
        grpc_fixture_data.units = self.units
        grpc_fixture_data.side = self.side
        grpc_fixture_data.height = self.height
        grpc_fixture_data.material = self.material
        grpc_fixture_data.state = self.state
        grpc_fixture_data.shape = self.shape
        grpc_fixture_data.x = self.x
        grpc_fixture_data.y = self.y
        grpc_fixture_data.length = self.length
        grpc_fixture_data.width = self.width
        grpc_fixture_data.diameter = self.diameter
        grpc_fixture_data.nodes = self.nodes
        grpc_fixture_data.rotation = self.rotation
        grpc_fixture_data.polygon = self.polygon
        grpc_fixture_data.boundary = self.boundary
        grpc_fixture_data.constraints = self.constraints
        grpc_fixture_data.chassisMaterial = self.chassis_material

        return grpc_fixture_data

    @field_validator("type", "units", "side", "state", "shape")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)


class UpdateICTFixturesRequest(BaseModel):
    """Contains the properties of an ict fixtures update per project."""

    project: str
    """Name of the Sherlock project."""
    cca_name: str
    """Name of the Sherlock CCA."""
    update_fixtures: list[ICTFixtureProperties]
    """List of ict fixtures with their properties to update"""

    @field_validator("project", "cca_name")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.UpdateICTFixturesRequest:
        request = SherlockLayerService_pb2.UpdateICTFixturesRequest()
        request.project = self.project
        request.ccaName = self.cca_name
        if self.update_fixtures is not None:
            for update_fixture in self.update_fixtures:
                request.ICTFixtureProperties.append(update_fixture._convert_to_grpc())
        return request


class MountPointProperties(BaseModel):
    """Contains the properties of a mount point."""

    id: str
    """ID"""
    type: str
    """Type"""
    shape: str
    """Shape type"""
    units: str
    """Units"""
    x: float
    """Center X"""
    y: float
    """Center Y"""
    length: float
    """Length"""
    width: float
    """Width"""
    diameter: float
    """Diameter"""
    nodes: str
    """Number of nodes"""
    rotation: float
    """Degrees of rotation"""
    side: str
    """Side"""
    height: float
    """Height"""
    material: str
    """Material"""
    boundary: str
    """Boundary point(s)"""
    constraints: str
    """FEA constraints"""
    polygon: str
    """Coordinates of points"""
    state: str
    """State"""
    chassis_material: str
    """Chassis material"""

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.MountPointProperties:
        grpc_mount_point_data = SherlockLayerService_pb2.MountPointProperties()

        grpc_mount_point_data.ID = self.id
        grpc_mount_point_data.type = self.type
        grpc_mount_point_data.shape = self.shape
        grpc_mount_point_data.units = self.units
        grpc_mount_point_data.x = str(self.x)
        grpc_mount_point_data.y = str(self.y)
        grpc_mount_point_data.length = str(self.length)
        grpc_mount_point_data.width = str(self.width)
        grpc_mount_point_data.diameter = str(self.diameter)
        grpc_mount_point_data.nodes = self.nodes
        grpc_mount_point_data.rotation = str(self.rotation)
        grpc_mount_point_data.side = self.side
        grpc_mount_point_data.height = str(self.height)
        grpc_mount_point_data.material = self.material
        grpc_mount_point_data.boundary = self.boundary
        grpc_mount_point_data.constraints = self.constraints
        grpc_mount_point_data.polygon = self.polygon
        grpc_mount_point_data.state = self.state
        grpc_mount_point_data.chassisMaterial = self.chassis_material

        return grpc_mount_point_data

    @field_validator("type", "shape", "units", "side", "state")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)


class GetMountPointsPropertiesRequest(BaseModel):
    """Return the properties for each mount point given a comma-separated list of mount point ids."""  # noqa: E501

    project: str
    """Name of the project."""
    cca_name: str
    """Name of the CCA containing the mount point properties to return."""
    mount_point_ids: Optional[str] = None
    """Optional Param: Comma-separated list of mount point ids representing one or more mount
        points. If this parameter is not included, then the entire list of mount points
        for a given CCA will have their properties returned.
    """

    @field_validator("project", "cca_name", "mount_point_ids")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    @field_validator("mount_point_ids")
    @classmethod
    def optional_str_validation(cls, value: Optional[str], info):
        """Allow the mount_point_ids to not be set, i.e., None."""
        return optional_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.GetMountPointsPropertiesRequest:
        request = SherlockLayerService_pb2.GetMountPointsPropertiesRequest()
        request.project = self.project
        request.ccaName = self.cca_name
        if self.mount_point_ids is not None:
            request.mountPointIDs = self.mount_point_ids
        return request


class UpdateMountPointsRequest(BaseModel):
    """Contains the properties of a mount point update per project."""

    project: str
    """Name of the Sherlock project."""
    cca_name: str
    """Name of the Sherlock CCA."""
    mount_points: list[MountPointProperties]
    """List of mount points with their properties to update"""

    @field_validator("project", "cca_name")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    @field_validator("mount_points")
    @classmethod
    def list_validation(cls, value: list, info: ValidationInfo):
        """Validate that mount_points is not empty."""
        if not value:
            raise ValueError(f"{info.field_name} must contain at least one item.")
        return value

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.UpdateMountPointsRequest:
        request = SherlockLayerService_pb2.UpdateMountPointsRequest()
        request.project = self.project
        request.ccaName = self.cca_name
        if self.mount_points is not None:
            for mount_point in self.mount_points:
                request.mountPointsProperties.append(mount_point._convert_to_grpc())
        else:
            raise ValueError("mount_points is invalid because it is None or empty.")

        return request



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing all stackup management capabilities."""
from typing import Optional

from grpc import Channel

try:
    import SherlockStackupService_pb2
    import SherlockStackupService_pb2_grpc
except ModuleNotFoundError:
    from ansys.api.sherlock.v0 import SherlockStackupService_pb2
    from ansys.api.sherlock.v0 import SherlockStackupService_pb2_grpc

from ansys.sherlock.core import LOG
from ansys.sherlock.core.errors import (
    SherlockGenStackupError,
    SherlockGetLayerCountError,
    SherlockGetStackupPropsError,
    SherlockGetTotalConductorThicknessError,
    SherlockInvalidConductorPercentError,
    SherlockInvalidGlassConstructionError,
    SherlockInvalidLayerIDError,
    SherlockInvalidMaterialError,
    SherlockListConductorLayersError,
    SherlockListLaminateLayersError,
    SherlockNoGrpcConnectionException,
    SherlockUpdateConductorLayerError,
    SherlockUpdateLaminateLayerError,
)
from ansys.sherlock.core.grpc_stub import GrpcStub
from ansys.sherlock.core.types.stackup_types import StackupProperties
from ansys.sherlock.core.utils.version_check import require_version


class Stackup(GrpcStub):
    """Contains all stackup management capabilities."""

    def __init__(self, channel: Channel, server_version: int):
        """Initialize a gRPC stub for Sherlock Stackup service."""
        super().__init__(channel, server_version)
        self.stub = SherlockStackupService_pb2_grpc.SherlockStackupServiceStub(channel)
        self.LAMINATE_THICKNESS_UNIT_LIST = None
        self.LAMINATE_MATERIAL_MANUFACTURER_LIST = None
        self.CONDUCTOR_MATERIAL_LIST = None
        self.LAYER_TYPE_LIST = ["SIGNAL", "POWER", "SUBSTRATE"]
        self.CONSTRUCTION_STYLE_LIST = None
        self.FIBER_MATERIAL_LIST = None

    def _init_laminate_thickness_units(self):
        """Initialize the list of units for the laminate thickness.

        .. deprecated:: 2026 R1

        """
        if self._is_connection_up():
            laminate_thickness_unit_request = (
                SherlockStackupService_pb2.ListLaminateThicknessUnitsRequest()
            )
            laminate_thickness_unit_response = self.stub.listLaminateThicknessUnits(
                laminate_thickness_unit_request
            )
            if laminate_thickness_unit_response.returnCode.value == 0:
                self.LAMINATE_THICKNESS_UNIT_LIST = laminate_thickness_unit_response.unit

    def _init_laminate_material_manufacturers(self):
        """Initialize list of laminate material manufacturers.

        Available Since: 2021R1
        """
        if self._is_connection_up():
            laminate_material_manufacturer_request = (
                SherlockStackupService_pb2.ListLaminateMaterialsManufacturersRequest()
            )
            laminate_material_manufacturer_response = self.stub.listLaminateMaterialsManufacturers(
                laminate_material_manufacturer_request
            )
            if laminate_material_manufacturer_response.returnCode.value == 0:
                self.LAMINATE_MATERIAL_MANUFACTURER_LIST = (
                    laminate_material_manufacturer_response.manufacturer
                )

    def _init_conductor_materials(self):
        """Initialize list of conductor materials.

        Available Since: 2021R1
        """
        if self._is_connection_up():
            conductor_materials_request = SherlockStackupService_pb2.ListConductorMaterialsRequest()
            conductor_materials_response = self.stub.listConductorMaterials(
                conductor_materials_request
            )
            if conductor_materials_response.returnCode.value == 0:
                self.CONDUCTOR_MATERIAL_LIST = conductor_materials_response.conductorMaterial

    def _init_construction_styles(self):
        """Initialize list of construction styles.

        Available Since: 2021R1
        """
        if self._is_connection_up():
            construction_style_request = SherlockStackupService_pb2.ListConstructionStylesRequest()
            construction_style_response = self.stub.listConstructionStyles(
                construction_style_request
            )
            if construction_style_response.returnCode.value == 0:
                self.CONSTRUCTION_STYLE_LIST = construction_style_response.constructionStyle

    def _init_fiber_materials(self):
        """Initialize list of fiber materials.

        Available Since: 2021R1
        """
        if self._is_connection_up():
            fiber_material_request = SherlockStackupService_pb2.ListFiberMaterialsRequest()
            fiber_material_response = self.stub.listFiberMaterials(fiber_material_request)
            if fiber_material_response.returnCode.value == 0:
                self.FIBER_MATERIAL_LIST = fiber_material_response.fiberMaterial

    def _check_pcb_material_validity(self, manufacturer: str, grade: str, material: str):
        """Check PCB arguments to see if they are valid.

        Available Since: 2021R1
        """
        if (self.LAMINATE_MATERIAL_MANUFACTURER_LIST is not None) and (
            manufacturer not in self.LAMINATE_MATERIAL_MANUFACTURER_LIST
        ):
            raise SherlockInvalidMaterialError(message="Laminate manufacturer is invalid.")
        if self._is_connection_up():
            request = SherlockStackupService_pb2.ListLaminateMaterialsRequest(
                manufacturer=manufacturer
            )
            response = self.stub.listLaminateMaterials(request)
            if response.returnCode.value == 0:
                manufacturer_materials = response.manufacturerMaterials[0]
                if manufacturer_materials.manufacturer != manufacturer:
                    raise SherlockInvalidMaterialError(
                        message=(
                            f"Unexpected manufacturer returned: "
                            f"{manufacturer_materials.manufacturer}"
                        )
                    )

                for grade_material in manufacturer_materials.gradeMaterials:
                    if grade == grade_material.grade:
                        if material in grade_material.laminateMaterial:
                            return

                        raise SherlockInvalidMaterialError("Laminate material is invalid.")
                else:
                    raise SherlockInvalidMaterialError("Laminate grade is invalid.")

    @staticmethod
    def _check_glass_construction_validity(glass_construction: list[tuple[str, float, float, str]]):
        """Check input to see if it is a valid glass construction argument."""
        if not isinstance(glass_construction, list):
            raise SherlockInvalidGlassConstructionError(
                message="glass_construction argument is invalid."
            )

        i = 0
        for i, layer in enumerate(glass_construction):
            if len(layer) != 4:
                error = (
                    f"Invalid glass construction layer {i}: "
                    "Number of elements in parameter is wrong."
                )
                raise SherlockInvalidGlassConstructionError(message=error)

    @staticmethod
    def _add_glass_construction_layers(
        request: SherlockStackupService_pb2.UpdateLaminateRequest,
        layers: list[tuple[str, float, float, str]],
    ):
        """Add the layers to the request."""
        for l in layers:
            layer = request.glassConstruction.add()
            layer.style = l[0]
            layer.resinPercentage = l[1]
            layer.thickness = l[2]
            layer.thicknessUnit = l[3]

    @require_version()
    def gen_stackup(
        self,
        project: str,
        cca_name: str,
        board_thickness: float,
        board_thickness_unit: str,
        pcb_material_manufacturer: str,
        pcb_material_grade: str,
        pcb_material: str,
        conductor_layers_cnt: int,
        signal_layer_thickness: float,
        signal_layer_thickness_unit: str,
        min_laminate_thickness: float,
        min_laminate_thickness_unit: str,
        maintain_symmetry: bool,
        power_layer_thickness: float,
        power_layer_thickness_unit: str,
    ) -> int:
        """Generate a new stackup from given properties.

        Available Since: 2021R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        board_thickness: float
            Board thickness.
        board_thickness_unit: str
            Units for the board thickness.
        pcb_material_manufacturer: str
            Manufacturer for the PCB material.
        pcb_material_grade: str
            Grade for the PCB material.
        pcb_material: str
            Material for the PCB.
        conductor_layers_cnt: int32
            Number of conductor layers.
        signal_layer_thickness: float
            Signal layer thickness.
        signal_layer_thickness_unit: str
            Units for the signal layer thickness.
        min_laminate_thickness: float
            Minimum thickness of laminate layers.
        min_laminate_thickness_unit: str
            Units for the minimum thickness of laminate layers.
        maintain_symmetry: bool
            Whether to maintain symmetry.
        power_layer_thickness: float
            Power layer thickness.
        power_layer_thickness_unit: str
            Units for the power layer thickness.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
            cca_name="Card",
        )
        >>> sherlock.stackup.gen_stackup(
            "Test",
            "Card",
            82.6,
            "mil",
            "Generic",
            "FR-4",
            "Generic FR-4",
            6,
            0.5,
            "oz",
            1.0,
            "mil",
            False,
            1.0,
            "mil"
        )
        """
        if self.LAMINATE_THICKNESS_UNIT_LIST is None:
            self._init_laminate_thickness_units()
        if self.LAMINATE_MATERIAL_MANUFACTURER_LIST is None:
            self._init_laminate_material_manufacturers()

        try:
            if project == "":
                raise SherlockGenStackupError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockGenStackupError(message="CCA name is invalid.")
            self._check_pcb_material_validity(
                pcb_material_manufacturer, pcb_material_grade, pcb_material
            )
            if conductor_layers_cnt <= 1:
                raise SherlockGenStackupError(
                    message="Number of conductor layers must be greater than 1."
                )
            if signal_layer_thickness < 0:
                raise SherlockGenStackupError(message="Conductor thickness is invalid.")
        except SherlockGenStackupError as e:
            LOG.error(str(e))
            raise e
        except SherlockInvalidMaterialError as e:
            LOG.error(f"Generate stackup error: {str(e)}")
            raise SherlockGenStackupError(message=str(e))

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockStackupService_pb2.GenStackupRequest(
            project=project,
            ccaName=cca_name,
            boardThickness=board_thickness,
            boardThicknessUnit=board_thickness_unit,
            pcbMaterialManufacturer=pcb_material_manufacturer,
            pcbMaterialGrade=pcb_material_grade,
            pcbMaterial=pcb_material,
            conductorLayersCnt=conductor_layers_cnt,
            signalLayerThickness=signal_layer_thickness,
            signalLayerThicknessUnit=signal_layer_thickness_unit,
            minLaminateThickness=min_laminate_thickness,
            minLaminateThicknessUnit=min_laminate_thickness_unit,
            maintainSymmetry=maintain_symmetry,
            powerLayerThickness=power_layer_thickness,
            powerLayerThicknessUnit=power_layer_thickness_unit,
        )

        response = self.stub.genStackup(request)

        try:
            if response.value == -1:
                raise SherlockGenStackupError(response.message)

            LOG.info(response.message)
            return response.value
        except SherlockGenStackupError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def update_conductor_layer(
        self,
        project: str,
        cca_name: str,
        layer: str,
        layer_type: str = "",
        material: str = "",
        thickness: float = 0,
        thickness_unit: str = "",
        conductor_percent: str = "",
        resin_material: str = "",
    ) -> int:
        """Update a conductor layer with given properties.

        Available Since: 2021R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        layer: str
            Layer ID associated with the conductor layer.
        layer_type: str, optional
            Layer type. The default is ``""``. For example,
            ``"SIGNAL"``, ``"POWER"``, or ``"SUBSTRATE"``.
        material: str, optional
            Conductor material. The default is ``""``.
        thickness: float, optional
            Conductor layer thickness. The default is ``0``.
        thickness_unit: str, optional
            Units for the conductor layer thickness. The
            default is ``""``.
        conductor_percent: str, optional
            Conductor percentage. The default is ``""``.
        resin_material: str, optional
            Resin material. The default is ``""``.

        Note
        ----
        Using the default value for a property causes no changes for that property.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
            cca_name="Card",
        )
        >>> sherlock.stackup.update_conductor_layer(
            "Test",
            "Card",
            "3",
            "POWER",
            "COPPER",
            1.0,
            "oz",
            "94.2",
            "Generic FR-4 Generic FR-4"
        )
        """
        if self.LAMINATE_THICKNESS_UNIT_LIST is None:
            self._init_laminate_thickness_units()
        if self.CONDUCTOR_MATERIAL_LIST is None:
            self._init_conductor_materials()

        try:
            if project == "":
                raise SherlockUpdateConductorLayerError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockUpdateConductorLayerError(message="CCA name is invalid.")
            if (layer_type != "") and layer_type not in self.LAYER_TYPE_LIST:
                raise SherlockUpdateConductorLayerError(
                    message=(
                        "Conductor type is invalid. "
                        'Options are "SIGNAL", "POWER", and "SUBSTRATE".'
                    )
                )
            if material != "":
                if (self.CONDUCTOR_MATERIAL_LIST is not None) and (
                    material not in self.CONDUCTOR_MATERIAL_LIST
                ):
                    raise SherlockUpdateConductorLayerError(
                        message="Conductor material is invalid."
                    )
        except SherlockUpdateConductorLayerError as e:
            LOG.error(str(e))
            raise e
        except (
            SherlockInvalidLayerIDError,
            SherlockInvalidConductorPercentError,
        ) as e:
            LOG.error(f"Update conductor layer error: {str(e)}")
            raise SherlockUpdateConductorLayerError(message=str(e))

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockStackupService_pb2.UpdateConductorLayerRequest(
            project=project,
            ccaName=cca_name,
            layer=layer,
            type=layer_type,
            material=material,
            thickness=thickness,
            thicknessUnit=thickness_unit,
            conductorPercent=conductor_percent,
            resinMaterial=resin_material,
        )

        response = self.stub.updateConductorLayer(request)

        try:
            if response.value == -1:
                raise SherlockUpdateConductorLayerError(response.message)

            LOG.info(response.message)
            return response.value
        except SherlockUpdateConductorLayerError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def update_laminate_layer(
        self,
        project,
        cca_name,
        layer,
        manufacturer: str = "",
        grade: str = "",
        material: str = "",
        thickness: float = 0,
        thickness_unit: str = "",
        construction_style: str = "",
        glass_construction: Optional[list[tuple[str, float, float, str]]] = None,
        fiber_material: str = "",
        conductor_material: str = "",
        conductor_percent: str = "",
    ) -> int:
        """Update a laminate layer with given properties.

        Available Since: 2021R1

        Parameters
        ----------
        project : str
            Name of the Sherlock project.
        cca_name : str
            Name of the CCA.
        layer : str
            Layer ID associated with the laminate layer.
        manufacturer : str, optional
            Manufacturer of the material for the laminate layer.
            The default is ``""``. To update the material, the
            ``manufacturer``, ``grade``, and ``material`` parameters
            must be specified. When the ``manufacturer`` is specified,
            there are checks to ensure that the corresponding parameters
            are provided.
        grade : str, optional
            Material grade. The default is ``""``.
        material : str, optional
            Material name. The default is ``""``.
        thickness : float, optional
            Laminate thickness. The default is ``0``.
        thickness_unit : str, optional
            Units for the laminate thickness. The default is ``""``.
        construction_style : str, optional
            Construction style. The default is ``""``.
        glass_construction : list[tuple[str, float, float, str]], optional
            List representing a glass construction. This list consists
            of objects with these properties:

           - style : str
               Style of the glass construction.
           - resinPercentage : float
               Resin percentage.
           - thickness: float
               Thickness.
           - thicknessUnit: str
               Units for the thickness.

        fiber_material : str, optional
            Fiber material. The default is ``""``. This parameter is only
            updated for a glass construction.
        conductor_material : str, optional
            Conductor material. The default is ``""``.
        conductor_percent : str, optional
            Conductor percentage. The default is ``""``.

        Note
        ----
        Using the default value for a property causes no changes for that property.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
            cca_name="Card",
        )
        >>> sherlock.stackup.update_laminate_layer(
            "Test",
            "Card",
            "2",
            "Generic",
            "FR-4",
            "Generic FR-4",
            0.015,
            "in",
            "106",
            [
                ("106", 68.0, 0.015, "in")
            ],
            "E-GLASS",
            "COPPER",
            "0.0"
        )
        """
        if self.LAMINATE_THICKNESS_UNIT_LIST is None:
            self._init_laminate_thickness_units()
        if self.LAMINATE_MATERIAL_MANUFACTURER_LIST is None:
            self._init_laminate_material_manufacturers()
        if self.CONSTRUCTION_STYLE_LIST is None:
            self._init_construction_styles()
        if self.FIBER_MATERIAL_LIST is None:
            self._init_fiber_materials()
        if self.CONDUCTOR_MATERIAL_LIST is None:
            self._init_conductor_materials()
        if glass_construction is None:
            glass_construction = []

        try:
            if project == "":
                raise SherlockUpdateLaminateLayerError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockUpdateLaminateLayerError(message="CCA name is invalid.")
            if manufacturer != "":
                self._check_pcb_material_validity(manufacturer, grade, material)
            if construction_style != "":
                if (self.CONSTRUCTION_STYLE_LIST is not None) and (
                    construction_style not in self.CONSTRUCTION_STYLE_LIST
                ):
                    raise SherlockUpdateLaminateLayerError(message="Construction style is invalid.")
                self._check_glass_construction_validity(glass_construction)
            if fiber_material != "":
                if (self.FIBER_MATERIAL_LIST is not None) and (
                    fiber_material not in self.FIBER_MATERIAL_LIST
                ):
                    raise SherlockUpdateLaminateLayerError(message="Fiber material is invalid.")
            if conductor_material != "":
                if (self.CONDUCTOR_MATERIAL_LIST is not None) and (
                    conductor_material not in self.CONDUCTOR_MATERIAL_LIST
                ):
                    raise SherlockUpdateLaminateLayerError(message="Conductor material is invalid.")
        except SherlockUpdateLaminateLayerError as e:
            LOG.error(str(e))
            raise
        except (
            SherlockInvalidLayerIDError,
            SherlockInvalidMaterialError,
            SherlockInvalidConductorPercentError,
            SherlockInvalidGlassConstructionError,
        ) as e:
            LOG.error(f"Update laminate layer error: %s", {str(e)})
            raise SherlockUpdateLaminateLayerError(message=str(e))

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockStackupService_pb2.UpdateLaminateRequest(
            project=project,
            ccaName=cca_name,
            layer=layer,
            manufacturer=manufacturer,
            grade=grade,
            material=material,
            thickness=thickness,
            thicknessUnit=thickness_unit,
            constructionStyle=construction_style,
            fiberMaterial=fiber_material,
            conductorMaterial=conductor_material,
            conductorPercent=conductor_percent,
        )

        self._add_glass_construction_layers(request, glass_construction)

        response = self.stub.updateLaminate(request)

        try:
            if response.value == -1:
                raise SherlockUpdateLaminateLayerError(response.message)

            LOG.info(response.message)
            return response.value
        except SherlockUpdateLaminateLayerError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def list_conductor_layers(
        self, project: str
    ) -> list[SherlockStackupService_pb2.ListConductorLayersResponse.CCAConductorLayerProp]:
        """List CCA conductor layers.

        Available Since: 2021R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.

        Returns
        -------
        list[SherlockStackupService_pb2.ListConductorLayersResponse.CCAConductorLayerProp]
            The conductor layers of all CCAs in the project.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
            cca_name="Card",
        )
        >>> conductorLayers = sherlock.stackup.list_conductor_layers(project="Tutorial")
        >>> for layer in conductorLayers:
        >>>     properties = layer.conductorLayerProps
        >>>     for prop in properties:
        >>>     print(f"{prop}")
        """
        if self.LAMINATE_THICKNESS_UNIT_LIST is None:
            self._init_laminate_thickness_units()
        if self.CONDUCTOR_MATERIAL_LIST is None:
            self._init_conductor_materials()

        if project == "":
            raise SherlockListConductorLayersError(message="Project name is invalid.")

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        try:

            request = SherlockStackupService_pb2.ListConductorLayersRequest(project=project)
            response = self.stub.listConductorLayers(request)
            if response.returnCode.value == -1:
                raise SherlockListConductorLayersError(response.returnCode.message)

            return response.ccaConductorLayerProps

        except SherlockListConductorLayersError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def list_laminate_layers(
        self, project: str
    ) -> list[SherlockStackupService_pb2.ListLaminatesResponse.CCALaminateProp]:
        """List all laminate layers and their properties.

        Available Since: 2021R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.

        Returns
        -------
        list[SherlockStackupService_pb2.ListLaminatesResponse.CCALaminateProp]
            The laminate layers of all CCAs in the project.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
            cca_name="Card",
        )
        >>> laminateLayers = sherlock.stackup.list_laminate_layers(project="Tutorial")
        >>> for layer in laminateLayers:
        >>>     properties = layer.laminateProps
        >>>     for prop in properties:
        >>>     print(f"{prop}")
        """
        if self.LAMINATE_THICKNESS_UNIT_LIST is None:
            self._init_laminate_thickness_units()
        if self.LAMINATE_MATERIAL_MANUFACTURER_LIST is None:
            self._init_laminate_material_manufacturers()
        if self.CONSTRUCTION_STYLE_LIST is None:
            self._init_construction_styles()
        if self.FIBER_MATERIAL_LIST is None:
            self._init_fiber_materials()
        if self.CONDUCTOR_MATERIAL_LIST is None:
            self._init_conductor_materials()

        if project == "":
            raise SherlockListLaminateLayersError(message="Project name is invalid.")
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        try:
            request = SherlockStackupService_pb2.ListLaminatesRequest(project=project)
            response = self.stub.listLaminates(request)
            if response.returnCode.value == -1:
                raise SherlockListLaminateLayersError(response.returnCode.message)

            return response.ccaLaminateProps

        except SherlockListLaminateLayersError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def get_layer_count(self, project: str, cca_name: str) -> int:
        """Get the number of CCA layers in a stackup.

        Available Since: 2021R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.

        Returns
        -------
        int
            The number of layers of the CCA in the project.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
            cca_name="Card",
        )
        >>> conductor_layer_count = sherlock.stackup.get_layer_count(
        >>>    project="Test",
        >>>    cca_name="Card")
        >>> print(f"{conductor_layer_count}")
        """
        if project == "":
            raise SherlockGetLayerCountError(message="Project name is invalid.")
        if cca_name == "":
            raise SherlockGetLayerCountError(message="CCA name is invalid.")
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockStackupService_pb2.GetLayerCountRequest(project=project, ccaName=cca_name)
        response = self.stub.getLayerCount(request)
        try:
            if response.returnCode.value == -1:
                raise SherlockGetLayerCountError(response.returnCode.message)

            return response.count
        except SherlockGetLayerCountError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def get_stackup_props(self, project: str, cca_name: str) -> StackupProperties:
        """Get the stackup properties from a CCA.

        Available Since: 2021R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.

        Returns
        -------
        StackupProperties
            Object containing the properties of the stackup.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
            cca_name="Card",
        )
        >>> stackup_props = sherlock.stackup.get_stackup_props(
               project="Tutorial",
               cca_name="Main Board"
            )
        >>> print(f"{stackup_props}")
        """
        if project == "":
            raise SherlockGetStackupPropsError(message="Project name is invalid.")
        if cca_name == "":
            raise SherlockGetStackupPropsError(message="CCA name is invalid.")
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockStackupService_pb2.GetStackupPropsRequest(
            project=project, ccaName=cca_name
        )
        response = self.stub.getStackupProps(request)
        try:
            if response.returnCode.value == -1:
                raise SherlockGetLayerCountError(response.returnCode.message)

            return StackupProperties(response)
        except SherlockGetStackupPropsError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def get_total_conductor_thickness(
        self, project: str, cca_name: str, thickness_unit: str
    ) -> float:
        """Return the total conductor thickness.

        Available Since: 2021R2

        Parameters
        ----------
        project: str
            Sherlock project name.
        cca_name: str
            The CCA name.
        thickness_unit: str, optional
            Units for laminate thickness.

        Returns
        -------
        float
            The conductor thickness of the CCA in the specified units.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
            cca_name="Card",
        )
        >>> total_thickness = sherlock.stackup.get_total_conductor_thickness(project="Tutorial",
                                                                 cca_name="Main Board",
                                                                 thickness_unit="oz")
        >>>print(f"{total_thickness}")
        """
        if project == "":
            raise SherlockGetTotalConductorThicknessError(message="Invalid project name")
        if cca_name == "":
            raise SherlockGetTotalConductorThicknessError(message="Invalid CCA name")
        if thickness_unit == "":
            raise SherlockGetTotalConductorThicknessError(message="Invalid thickness unit")
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockStackupService_pb2.GetTotalConductorThicknessRequest(
            project=project, ccaName=cca_name, thicknessUnit=thickness_unit
        )
        response = self.stub.getTotalConductorThickness(request)

        try:
            if response.returnCode.value == -1:
                raise SherlockGetTotalConductorThicknessError(response.returnCode.message)

            return response.totalThickness
        except SherlockGetTotalConductorThicknessError as e:
            LOG.error(str(e))
            raise e



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing types for the Parts Service."""

from typing import List, Optional
import warnings

from pydantic import BaseModel, field_validator

from ansys.sherlock.core.types.common_types import basic_str_validator

try:
    import SherlockCommonService_pb2
    import SherlockPartsService_pb2
except ModuleNotFoundError:
    from ansys.api.sherlock.v0 import SherlockCommonService_pb2, SherlockPartsService_pb2

parts_service = SherlockPartsService_pb2


def deprecation(cls: object):
    """Raise a DeprecationWarning when a deprecated class is used."""
    message = f"{cls} is deprecated. Use a string with the value of the constant name \
    as defined in the proto file."
    warnings.warn(message, DeprecationWarning, stacklevel=2)
    return cls


@deprecation
class PartsListSearchMatchingMode:
    """DEPRECATED. Constants for Matching Mode in Update Parts List & Update Parts from AVL."""

    matching_mode = SherlockCommonService_pb2.MatchingMode
    BOTH = matching_mode.Both
    """Both"""
    PART = matching_mode.Part
    """Part"""


class PartsListSearchDuplicationMode:
    """Constants for Duplication Mode in Update Parts List and Update Parts from AVL request."""

    duplication_mode = SherlockPartsService_pb2.DuplicationMode
    FIRST = duplication_mode.First
    """First"""
    ERROR = duplication_mode.Error
    """Error"""
    IGNORE = duplication_mode.Ignore
    """Ignore"""


class AVLPartNum:
    """Constants for AVLPartNum in the Update Parts List from AVL request."""

    avl_part_num = SherlockPartsService_pb2.AVLPartNum
    ASSIGN_INTERNAL_PART_NUM = avl_part_num.AssignInternalPartNum
    """AssignInternalPartNum"""
    ASSIGN_VENDOR_AND_PART_NUM = avl_part_num.AssignVendorAndPartNum
    """AssignVendorAndPartNum"""
    DO_NOT_CHANGE_VENDOR_OR_PART_NUM = avl_part_num.DoNotChangeVendorOrPartNum
    """DoNotChangeVendorOrPartNum"""


class AVLDescription:
    """Constants for AVLDescription in the Update Parts List from AVL request."""

    avl_description = SherlockPartsService_pb2.AVLDescription
    ASSIGN_APPROVED_DESCRIPTION = avl_description.AssignApprovedDescription
    """AssignApprovedDescription"""
    DO_NOT_CHANGE_DESCRIPTION = avl_description.DoNotChangeDescription
    """DoNotChangeDescription"""


class GetPartsListPropertiesRequest(BaseModel):
    """Request for getting properties of parts in the parts list of a CCA."""

    project: str
    """Name of the Sherlock project."""
    cca_name: str
    """Name of the CCA with the parts."""
    reference_designators: Optional[List[str]] = None
    """Reference designators of the parts to retrieve properties for. Use None to get all parts."""

    @field_validator("project", "cca_name")
    @classmethod
    def str_validation(cls, value: str, info):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> SherlockPartsService_pb2.GetPartsListPropertiesRequest:
        return SherlockPartsService_pb2.GetPartsListPropertiesRequest(
            project=self.project,
            ccaName=self.cca_name,
            refDes=self.reference_designators,
        )


class UpdatePadPropertiesRequest(BaseModel):
    """Contains the properties to update pad properties for one or more parts in a parts list."""

    project: str
    """Name of the Sherlock project."""
    cca_name: str
    """Name of the CCA for which pad properties will be updated."""
    reference_designators: Optional[List[str]] = None
    """Reference designators of the associated parts to be updated."""

    @field_validator("project", "cca_name")
    @classmethod
    def str_validation(cls, value: str, info):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> parts_service.UpdatePadPropertiesRequest:
        return parts_service.UpdatePadPropertiesRequest(
            project=self.project,
            ccaName=self.cca_name,
            refDes=self.reference_designators,
        )


class DeletePartsFromPartsListRequest(BaseModel):
    """Contains the information to delete parts from the parts list for a given project's CCA."""

    project: str
    """Name of the Sherlock project."""
    cca_name: str
    """Name of the CCA for which parts will be deleted."""
    reference_designators: Optional[List[str]] = None
    """Reference designators of the associated parts to be deleted."""

    @field_validator("project", "cca_name")
    @classmethod
    def str_validation(cls, value: str, info):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> parts_service.DeletePartsFromPartsListRequest:
        return parts_service.DeletePartsFromPartsListRequest(
            project=self.project,
            ccaName=self.cca_name,
            refDes=self.reference_designators,
        )


class ImportPartsToAVLRequest(BaseModel):
    """Request to import parts into the Approved Vendor List (AVL)."""

    import_file: str
    """Full file path to the AVL file."""

    import_type: parts_service.AVLImportType.ValueType
    """Import mode to use for AVL data."""

    """Allow non-standard types like Protobuf enums in Pydantic models."""
    model_config = {"arbitrary_types_allowed": True}

    @field_validator("import_file")
    @classmethod
    def import_file_validator(cls, value: str, info):
        """Validate that the import file path is not empty."""
        if value.strip() == "":
            raise ValueError(f"{info.field_name} cannot be empty.")
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> parts_service.ImportPartsToAVLRequest:
        request = parts_service.ImportPartsToAVLRequest()
        request.importFile = self.import_file
        request.importType = self.import_type
        return request



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing all parts management capabilities."""
try:
    import SherlockCommonService_pb2
    import SherlockPartsService_pb2
    import SherlockPartsService_pb2_grpc
except ModuleNotFoundError:
    from ansys.api.sherlock.v0 import SherlockCommonService_pb2
    from ansys.api.sherlock.v0 import SherlockPartsService_pb2
    from ansys.api.sherlock.v0 import SherlockPartsService_pb2_grpc

from ansys.sherlock.core import LOG
from ansys.sherlock.core.errors import (
    SherlockEnableLeadModelingError,
    SherlockExportNetListError,
    SherlockExportPartsListError,
    SherlockImportPartsListError,
    SherlockNoGrpcConnectionException,
    SherlockUpdatePartsFromAVLError,
    SherlockUpdatePartsListError,
    SherlockUpdatePartsListPropertiesError,
    SherlockUpdatePartsLocationsByFileError,
    SherlockUpdatePartsLocationsError,
)
from ansys.sherlock.core.grpc_stub import GrpcStub
from ansys.sherlock.core.types.common_types import TableDelimiter
from ansys.sherlock.core.types.parts_types import (
    AVLDescription,
    AVLPartNum,
    DeletePartsFromPartsListRequest,
    GetPartsListPropertiesRequest,
    ImportPartsToAVLRequest,
    PartsListSearchDuplicationMode,
    UpdatePadPropertiesRequest,
)
from ansys.sherlock.core.utils.version_check import require_version


class Parts(GrpcStub):
    """Contains all parts management capabilities."""

    def __init__(self, channel, server_version):
        """Initialize a gRPC stub for the Sherlock Parts service."""
        super().__init__(channel, server_version)
        self.stub = SherlockPartsService_pb2_grpc.SherlockPartsServiceStub(channel)
        self.PART_LOCATION_UNITS = None
        self.BOARD_SIDES = None

    @staticmethod
    def _add_part_loc_request(
        request: SherlockPartsService_pb2.UpdatePartsLocationsRequest,
        parts: list[tuple[str, str, str, str, str, str, str]],
    ) -> None:
        """Add part locations to the request."""
        for p in parts:
            part = request.partLoc.add()
            part.refDes = p[0]
            part.x = p[1]
            part.y = p[2]
            part.rotation = p[3]
            part.locationUnits = p[4]
            part.boardSide = p[5]
            part.mirrored = p[6]

    def _check_part_loc_validity(
        self, part_locations: list[tuple[str, str, str, str, str, str, str]]
    ) -> None:
        """Check input to see if it is a valid part location list."""
        if not isinstance(part_locations, list):
            raise SherlockUpdatePartsLocationsError(message="Part location argument is invalid.")

        if len(part_locations) == 0:
            raise SherlockUpdatePartsLocationsError(message="Part location properties are missing.")
        for i, part in enumerate(part_locations):
            if len(part) != 7:
                raise SherlockUpdatePartsLocationsError(
                    message=f"Invalid part location {i}: Number of fields is invalid."
                )
            if part[0] == "":
                raise SherlockUpdatePartsLocationsError(
                    message=f"Invalid part location {i}: Reference designator is missing."
                )
            if part[4] != "":
                if self.PART_LOCATION_UNITS is not None and part[4] not in self.PART_LOCATION_UNITS:
                    raise SherlockUpdatePartsLocationsError(
                        message=f"Invalid part location {i}: Location units are invalid."
                    )
            if part[1] != "":
                if part[4] == "":
                    raise SherlockUpdatePartsLocationsError(
                        message=f"Invalid part location {i}: Location units are missing."
                    )
                try:
                    float(part[1])
                except ValueError:
                    raise SherlockUpdatePartsLocationsError(
                        message=(
                            f"Invalid part location {i}: " f"Location X coordinate is invalid."
                        )
                    )
            if part[2] != "":
                if part[4] == "":
                    raise SherlockUpdatePartsLocationsError(
                        message=f"Invalid part location {i}: Location units are missing."
                    )
                try:
                    float(part[2])
                except ValueError:
                    raise SherlockUpdatePartsLocationsError(
                        message=(
                            f"Invalid part location {i}: " f"Location Y coordinate is invalid."
                        )
                    )
            if part[3] != "":
                try:
                    rotation = float(part[3])
                    if rotation < -360 or rotation > 360:
                        raise SherlockUpdatePartsLocationsError(
                            message=(
                                f"Invalid part location {i}: " f"Location rotation is invalid."
                            )
                        )
                except ValueError:
                    raise SherlockUpdatePartsLocationsError(
                        message=f"Invalid part location {i}: Location rotation is invalid."
                    )
            if part[5] != "":
                if self.BOARD_SIDES is not None and part[5] not in self.BOARD_SIDES:
                    raise SherlockUpdatePartsLocationsError(
                        message=f"Invalid part location {i}: Location board side is invalid."
                    )
            if part[6] != "":
                if part[6] != "True" and part[6] != "False":
                    raise SherlockUpdatePartsLocationsError(
                        message=f"Invalid part location {i}: Location mirrored is invalid."
                    )

    def _init_location_units(self):
        """Initialize units for part location.

        Available since: 2022R1
        """
        if self._is_connection_up():
            part_location_request = SherlockPartsService_pb2.GetPartLocationUnitsRequest()
            part_location_response = self.stub.getPartLocationUnits(part_location_request)
            if part_location_response.returnCode.value == 0:
                self.PART_LOCATION_UNITS = part_location_response.units

    def _init_board_sides(self):
        """Initialize board sides.

        Available Since: 2022R1
        """
        if self._is_connection_up():
            board_sides_request = SherlockPartsService_pb2.GetBoardSidesRequest()
            board_sides_response = self.stub.getBoardSides(board_sides_request)
            if board_sides_response.returnCode.value == 0:
                self.BOARD_SIDES = board_sides_response.boardSides

    @require_version()
    def update_parts_list(
        self,
        project: str,
        cca_name: str,
        part_library: str,
        matching_mode: str,
        duplication_mode: PartsListSearchDuplicationMode,
    ) -> int:
        """Update a parts list based on matching and duplication preferences.

        Available Since: 2021R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        part_library: str
            Name of the parts library.
        matching_mode: str
            Matching mode for updates.
        duplication_mode: PartsListSearchDuplicationMode
            How to handle duplication during the update.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> import SherlockCommonService_pb2
        >>> import SherlockPartsService_pb2
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.parts.update_parts_list(
        >>>     "Test",
        >>>     "Card",
        >>>     "Sherlock Part Library",
        >>>     SherlockCommonService_pb2.MatchingMode.Both,
        >>>     SherlockPartsService_pb2.DuplicationMode.Error,
        >>> )
        """
        try:
            if project == "":
                raise SherlockUpdatePartsListError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockUpdatePartsListError(message="CCA name is invalid.")
            if part_library == "":
                raise SherlockUpdatePartsListError(message="Parts library is invalid.")
        except SherlockUpdatePartsListError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockPartsService_pb2.UpdatePartsListRequest(
            project=project,
            ccaName=cca_name,
            partLibrary=part_library,
            matching=matching_mode,
            duplication=duplication_mode,
        )

        response = self.stub.updatePartsList(request)

        return_code = response.returnCode

        try:
            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockUpdatePartsListError(error_array=response.errors)

                raise SherlockUpdatePartsListError(message=return_code.message)
            else:
                LOG.info(return_code.message)
                return return_code.value
        except SherlockUpdatePartsListError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version()
    def update_parts_locations(
        self, project: str, cca_name: str, part_loc: list[tuple[str, str, str, str, str, str, str]]
    ) -> int:
        """Update one or more part locations.

        Available Since: 2022R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        part_loc: list[tuple[str, str, str, str, str, str, str]]
            List defining the part locations. The list consists
            of these properties:

            - refDes: str
                Reference designator of the part.
            - x: str
                Value for the x coordinate.
            - y: str
                Value for the y coordinate.
            - rotation: str
                Rotation.
            - location_units: str
                Locations units.
            - board_side: str
                Board side.
            - mirrored: str
                Mirrored.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.parts.update_parts_locations(
        >>>     "Test",
        >>>     "Card",
        >>>     [
        >>>         ("C1", "-2.7", "-1.65", "0", "in", "TOP", "False"),
        >>>         ("J1", "-3.55", "-2.220446049250313E-16", "90", "in", "TOP", "False"),
        >>>     ]
        >>> )
        """
        if self.PART_LOCATION_UNITS is None:
            self._init_location_units()
        if self.BOARD_SIDES is None:
            self._init_board_sides()

        try:
            if project == "":
                raise SherlockUpdatePartsLocationsError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockUpdatePartsLocationsError(message="CCA name is invalid.")
            self._check_part_loc_validity(part_loc)
        except SherlockUpdatePartsLocationsError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockPartsService_pb2.UpdatePartsLocationsRequest(
            project=project,
            ccaName=cca_name,
        )

        self._add_part_loc_request(request, part_loc)

        response = self.stub.updatePartsLocations(request)

        return_code = response.returnCode

        try:
            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockUpdatePartsLocationsError(error_array=response.updateError)

                raise SherlockUpdatePartsLocationsError(message=return_code.message)
            else:
                LOG.info(return_code.message)
                return return_code.value
        except SherlockUpdatePartsLocationsError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version()
    def update_parts_locations_by_file(
        self, project: str, cca_name: str, file_path: str, numeric_format: str = ""
    ) -> int:
        """Update one or more part locations using a CSV file.

        Available Since: 2023R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        file_path: str
            Full path to the file with the components and location properties.
        numeric_format: str, optional
            Numeric format for the file, which indicates whether commas or points
            are used as decimal markers. The default is ``""``, in which case
            ``"English (United States)"`` is the numeric format. This
            indicates that points are used as decimal markers.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.parts.update_parts_locations_by_file(
        >>>     "Test",
        >>>     "Card",
        >>>     "Parts Locations.csv"
        >>> )
        """
        try:
            if project == "":
                raise SherlockUpdatePartsLocationsByFileError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockUpdatePartsLocationsByFileError(message="CCA name is invalid.")
            if file_path == "":
                raise SherlockUpdatePartsLocationsByFileError(message="Filepath is required.")
        except SherlockUpdatePartsLocationsByFileError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockPartsService_pb2.UpdatePartsLocationsByFileRequest(
            project=project,
            ccaName=cca_name,
            numericFormat=numeric_format,
            filePath=file_path,
        )

        response = self.stub.updatePartsLocationsByFile(request)

        return_code = response.returnCode

        try:
            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockUpdatePartsLocationsByFileError(error_array=response.updateError)

                raise SherlockUpdatePartsLocationsByFileError(message=return_code.message)
            else:
                LOG.info(return_code.message)
                return return_code.value
        except SherlockUpdatePartsLocationsByFileError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version()
    def import_parts_list(
        self, project: str, cca_name: str, import_file: str, import_as_user_src: bool
    ) -> int:
        """Import a parts list for a CCA.

        Available Since: 2021R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        import_file: str
            Full path to the CSV file with the parts list.
        import_as_user_src: bool
            Whether to set the data source of the properties to ``"User"``.
            Otherwise, the data source is set to the name of the CSV file.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.parts.import_parts_list(
        >>>     "Test",
        >>>     "Card",
        >>>     "Parts List.csv",
        >>>     False
        >>> )
        """
        try:
            if project == "":
                raise SherlockImportPartsListError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockImportPartsListError(message="CCA name is invalid.")
            if import_file == "":
                raise SherlockImportPartsListError(message="Import filepath is required.")
        except SherlockImportPartsListError as e:
            LOG.error(str(e))
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockPartsService_pb2.ImportPartsListRequest(
            project=project,
            ccaName=cca_name,
            importFile=import_file,
            importAsUserSrc=import_as_user_src,
        )

        response = self.stub.importPartsList(request)

        try:
            if response.value == -1:
                raise SherlockImportPartsListError(response.message)

            LOG.info(response.message)
            return response.value
        except SherlockImportPartsListError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def export_parts_list(self, project: str, cca_name: str, export_file: str) -> int:
        """Export a parts list for a CCA.

        Available Since: 2021R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        export_file: str
            Full path for the CSV file to export the parts list to.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.parts.export_parts_list(
        >>>     "Test",
        >>>     "Card",
        >>>     "Parts List.csv"
        >>> )
        """
        try:
            if project == "":
                raise SherlockExportPartsListError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockExportPartsListError(message="CCA name is invalid.")
            if export_file == "":
                raise SherlockExportPartsListError(message="Export filepath is required.")
        except SherlockExportPartsListError as e:
            LOG.error(str(e))
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockPartsService_pb2.ExportPartsListRequest(
            project=project, ccaName=cca_name, exportFile=export_file
        )

        response = self.stub.exportPartsList(request)

        try:
            if response.value == -1:
                raise SherlockExportPartsListError(response.message)

            LOG.info(response.message)
            return response.value
        except SherlockExportPartsListError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def enable_lead_modeling(self, project: str, cca_name: str):
        """Enable lead modeling for leaded parts.

        Available Since: 2021R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.parts.enable_lead_modeling(
        >>>     "Test",
        >>>     "Card"
        >>> )
        """
        try:
            if project == "":
                raise SherlockEnableLeadModelingError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockEnableLeadModelingError(message="CCA name is invalid.")
        except SherlockEnableLeadModelingError as e:
            LOG.error(str(e))
            raise e

        request = SherlockPartsService_pb2.UpdateLeadModelingRequest(
            project=project,
            ccaName=cca_name,
        )

        response = self.stub.updateLeadModeling(request)

        try:
            if response.value == -1:
                raise SherlockEnableLeadModelingError(response.message)

            LOG.info(response.message)
            return response.value
        except SherlockEnableLeadModelingError as e:
            LOG.error(str(e))
            raise e

    @require_version(252)
    def get_parts_list_properties(
        self, request: GetPartsListPropertiesRequest
    ) -> list[SherlockPartsService_pb2.GetPartsListPropertiesResponse]:
        """Return the properties for one or more parts in the parts list for the CCA.

        Available Since: 2025R2

        Parameters
        ----------
        request: GetPartsListPropertiesRequest
            Contains the information needed to retrieve the properties of parts in the parts list.

        Returns
        -------
        list[SherlockPartsService_pb2.GetPartsListPropertiesResponse]
            Properties for each part that corresponds to the reference designators.

        Examples
        --------
        >>> from ansys.sherlock.core.types.parts_types import (GetPartsListPropertiesRequest)
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>    "ODB++ Tutorial.tgz",
        >>>    True,
        >>>    True,
        >>>    True,
        >>>    True,
        >>>    project="Test",
        >>>    cca_name="Card",
        >>> )
        >>> part_properties = sherlock.parts.get_parts_list_properties(
        >>>     GetPartsListPropertiesRequest(
        >>>         project="Test",
        >>>         cca_name="Card",
        >>>         reference_designators=["C1","U9"]
        >>>     )
        >>> )
        >>> print(f"{part_properties}")
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        return list(self.stub.getPartsListProperties(request._convert_to_grpc()))

    @require_version(241)
    def update_parts_from_AVL(
        self,
        project: str,
        cca_name: str,
        matching_mode: str,
        duplication_mode: PartsListSearchDuplicationMode,
        avl_part_num: AVLPartNum,
        avl_description: AVLDescription,
    ) -> SherlockPartsService_pb2.UpdatePartsListFromAVLResponse:
        r"""Update the parts list from the Approved Vendor List (AVL).

        Available Since: 2024R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        matching_mode: str
            Determines how parts are matched against the AVL
        duplication_mode: PartsListSearchDuplicationMode
            Determines how duplicate part matches are handled when found
        avl_part_num: AVLPartNum
            Determines what part number info in the parts list is updated from the AVL
        avl_description: AVLDescription
            Determines if the part description is updated or not

        Returns
        -------
        UpdatePartsListFromAVLResponse
            - returnCode: ReturnCode
                - value: int
                    Status code of the response. 0 for success.
                - message: str
                    Indicates general errors that occurred while attempting to update parts
            - numPartsUpdated: int
                Number of parts updated
            - updateErrors: list<str>
                Errors found when updating part

        Examples
        --------
        >>> from ansys.sherlock.core.types.parts_types import (
            AVLDescription,
            AVLPartNum,
            PartsListSearchDuplicationMode,
        )
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "C:\\Program Files\\ANSYS Inc\\v241\\sherlock\\tutorial\\ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.parts.update_parts_from_AVL(
        >>>     project="Test",
        >>>     cca_name="Card",
        >>>     matching_mode="Both",
        >>>     duplication=PartsListSearchDuplicationMode.FIRST,
        >>>     avl_part_num=AVLPartNum.ASSIGN_INTERNAL_PART_NUM,
        >>>     avl_description=AVLDescription.ASSIGN_APPROVED_DESCRIPTION
        >>> )
        """
        try:
            if project == "":
                raise SherlockUpdatePartsFromAVLError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockUpdatePartsFromAVLError(message="CCA name is invalid.")

            request = SherlockPartsService_pb2.UpdatePartsListFromAVLRequest(
                project=project,
                ccaName=cca_name,
                matching=matching_mode,
                duplication=duplication_mode,
                avlPartNum=avl_part_num,
                avlDesc=avl_description,
            )

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            # Call method on server
            response = self.stub.updatePartsListFromAVL(request)

            return_code = response.returnCode

            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockUpdatePartsFromAVLError(error_array=response.updateErrors)
                raise SherlockUpdatePartsFromAVLError(message=return_code.message)

            return return_code.value
        except SherlockUpdatePartsFromAVLError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version(242)
    def update_parts_list_properties(
        self,
        project: str,
        cca_name: str,
        part_properties: list[dict[str, list[dict[str, str]] | list[str]]],
    ) -> int:
        """
        Update one or more properties of one or more parts in a parts list.

        Available Since: 2024R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        part_properties: list[dict[str, list[str] | list[dict[str, str]]]]
            Part properties consisting of these properties:

                - reference_designators: list[str], optional
                    Reference designator for each part to be updated. If not included,
                    update properties for all parts in the CCA.
                - properties: list[dict[str, str]]
                    Part properties consisting of these properties:

                        - name: str
                            Name of property to be updated.
                        - value: str
                            Value to be applied to the chosen part property.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.parts.update_parts_list_properties(
        >>>     "Test",
        >>>     "Card",
        >>>     [
        >>>         {
        >>>             "reference_designators": ["C1"],
        >>>             "properties": [
        >>>                 {"name": "partType", "value": "RESISTOR"}
        >>>             ]
        >>>         },
        >>>         {
        >>>             "reference_designators": ["C2"],
        >>>             "properties": [
        >>>                 {"name": "locX", "value": "1"}
        >>>             ]
        >>>         }
        >>>     ]
        >>> )
        """
        try:
            if project == "":
                raise SherlockUpdatePartsListPropertiesError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockUpdatePartsListPropertiesError(message="CCA name is invalid.")
            if len(part_properties) == 0:
                raise SherlockUpdatePartsListPropertiesError(message="Part properties are missing.")

            for i, part_property in enumerate(part_properties):
                if len(part_property) < 1 or len(part_property) > 2:
                    raise SherlockUpdatePartsListPropertiesError(
                        f"Number of elements ({len(part_property)}) "
                        f"is wrong for part list property {i}."
                    )
                elif not isinstance(part_property["reference_designators"], list):
                    raise SherlockUpdatePartsListPropertiesError(
                        f"reference_designators is not a list " f"for parts list property {i}."
                    )

                properties = part_property["properties"]
                for j, prop in enumerate(properties):
                    if len(prop) < 1 or len(prop) > 2:
                        raise SherlockUpdatePartsListPropertiesError(
                            f"Number of elements ({len(prop)}) " f"is wrong for property {j}."
                        )
                    elif not isinstance(prop["name"], str) or prop["name"] == "":
                        raise SherlockUpdatePartsListPropertiesError(
                            f"Name is required " f"for property {j}."
                        )
                    elif not isinstance(prop["value"], str):
                        raise SherlockUpdatePartsListPropertiesError(message="Value is invalid.")

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockPartsService_pb2.UpdatePartsListPropertiesRequest(
                project=project, ccaName=cca_name
            )

            # Add part properties to the request
            for part_prop in part_properties:
                prop = request.partProperties.add()
                reference_designators = part_prop["reference_designators"]
                if reference_designators is not None:
                    for ref_des in reference_designators:
                        prop.refDes.append(ref_des)

                props = part_prop["properties"]
                if props is not None:
                    for prop_dict in props:
                        property_obj = prop.properties.add()
                        property_obj.name = prop_dict["name"]
                        property_obj.value = prop_dict["value"]

            response = self.stub.updatePartsListProperties(request)

            return_code = response.returnCode

            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockUpdatePartsListPropertiesError(error_array=response.errors)

                raise SherlockUpdatePartsListPropertiesError(message=return_code.message)

            return return_code.value

        except SherlockUpdatePartsListPropertiesError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version(242)
    def export_net_list(
        self,
        project: str,
        cca_name: str,
        output_file: str,
        col_delimiter: str = TableDelimiter.COMMA,
        overwrite_existing: bool = False,
        utf8_enabled: bool = False,
    ) -> int:
        """Export a net list to a delimited output file.

        Available Since: 2024R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        output_file: str
            Full path for the output file where the net list will be written.
        col_delimiter: TableDelimiter, optional
            The delimiter character to be used. Defaults to TableDelimiter.COMMA.
        overwrite_existing: bool, optional
            Flag to determine if existing .CSV files should be overwritten
            if they match the output_file. Defaults to False.
        utf8_enabled: bool, optional
            Flag that specifies if UTF-8 will be used for .CSV files. Defaults to False.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.parts.export_net_list(
        >>>     "Test",
        >>>     "Card",
        >>>     "Net List.csv",
        >>>     col_delimiter=TableDelimiter.TAB,
        >>>     overwrite_existing=True,
        >>>     utf8_enabled=True
        >>> )
        """
        try:
            if project == "":
                raise SherlockExportNetListError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockExportNetListError(message="CCA name is invalid.")
            if output_file == "":
                raise SherlockExportNetListError(message="Output file path is required.")

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockPartsService_pb2.ExportNetListRequest(
                project=project,
                ccaName=cca_name,
                outputFilePath=output_file,
                overwriteExisting=overwrite_existing,
                colDelimiter=col_delimiter,
                utf8Enabled=utf8_enabled,
            )

            response = self.stub.exportNetList(request)

            if response.value == -1:
                raise SherlockExportNetListError(response.message)

        except SherlockExportNetListError as e:
            LOG.error(str(e))
            raise e

        return response.value

    @require_version(252)
    def update_pad_properties(
        self,
        request: UpdatePadPropertiesRequest,
    ) -> list[SherlockPartsService_pb2.UpdatePadPropertiesResponse]:
        r"""Update pad properties for one or more parts in a parts list.

        Parameters
        ----------
        request: UpdatePadPropertiesRequest
            Contains all the information needed to update the pad properties for one or more parts
            in a project's CCA.

        Returns
        -------
        list[SherlockPartsService_pb2.UpdatePadPropertiesResponse]
            Status of the pad properties update and any error messages.

        Examples
        --------
        >>> from ansys.sherlock.core.types.parts_types import (
        >>>     UpdatePadPropertiesRequest,
        >>> )
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_project_zip_archive(
        >>>     project="Assembly Tutorial",
        >>>     category="category",
        >>>     archive_file=\
        >>>         "C:\\Program Files\\ANSYS Inc\\v252\\sherlock\\tutorial\\Assembly Tutorial.zip",
        >>> )
        >>> request = UpdatePadPropertiesRequest(
        >>>     project="Assembly Tutorial",
        >>>     cca_name="Main Board",
        >>>     reference_designators=["U1", "R2", "C3"]
        >>> )
        >>> response = sherlock.parts.update_pad_properties(request)
        >>> for res in response:
        >>>     print(f"Return code: value={res.returnCode.value},\
        >>>     message={res.returnCode.message},\
        >>>     reference_designators={res.reference_designators}")
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        update_request = request._convert_to_grpc()

        return list(self.stub.updatePadProperties(update_request))

    @require_version(252)
    def delete_parts_from_parts_list(
        self,
        request: DeletePartsFromPartsListRequest,
    ) -> list[SherlockPartsService_pb2.DeletePartsFromPartsListResponse]:
        r"""Delete parts from the parts list for a given project's CCA.

        Parameters
        ----------
        request: DeletePartsFromPartsListRequest
            Contains all the information needed to delete parts from the
            parts list in a project's CCA.

        Returns
        -------
        list[SherlockPartsService_pb2.DeletePartsFromPartsListResponse]
            Status of the delete operation and any error messages.

        Examples
        --------
        >>> from ansys.sherlock.core.types.parts_types import (
        >>>     DeletePartsFromPartsListRequest,
        >>> )
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_project_zip_archive(
        >>>     project="Assembly Tutorial",
        >>>     category="category",
        >>>     archive_file=\
        >>>         "C:\\Program Files\\ANSYS Inc\\v252\\sherlock\\tutorial\\Assembly Tutorial.zip",
        >>> )
        >>> request = DeletePartsFromPartsListRequest(
        >>>     project="Assembly Tutorial",
        >>>     cca_name="Main Board",
        >>>     reference_designators=["U1", "R2", "C3"],
        >>> )
        >>> response = sherlock.parts.delete_parts_from_parts_list(request)
        >>> for res in response:
        >>>     print(f"Return code: value={res.returnCode.value},\
        >>>     message={res.returnCode.message},\
        >>>     reference designator={res.reference_designators}")
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        delete_request = request._convert_to_grpc()

        return list(self.stub.deletePartsFromPartsList(delete_request))

    @require_version(261)
    def import_parts_to_avl(
        self,
        request: ImportPartsToAVLRequest,
    ) -> SherlockCommonService_pb2.ReturnCode:
        """Import a parts list into the Approved Vendor List (AVL).

        Available Since: 2026R1

        Parameters
        ----------
        request : ImportPartsToAVLRequest
            Contains the file path and import mode to use for the AVL parts import.

        Returns
        -------
        SherlockCommonService_pb2.ReturnCode
            Return code indicating the result of the AVL parts import.

        Examples
        --------
        >>> from ansys.sherlock.core.types.part_types import ImportPartsToAVLRequest
        >>> from ansys.api.sherlock.v0 import SherlockPartsService_pb2
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> return_code = sherlock.project.import_parts_to_avl(
        >>>     ImportPartsToAVLRequest(
        >>>         import_file="C:/path/to/AVL_parts_file.xls",
        >>>         import_type=SherlockPartsService_pb2.AVLImportType.Update
        >>>     )
        >>> )
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        return self.stub.importPartsToAVL(request._convert_to_grpc())



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.

"""Module containing types for the Lifecycle Service."""

from typing import Optional

from pydantic import BaseModel, ValidationInfo, field_validator

from ansys.sherlock.core.types.common_types import basic_str_validator, optional_str_validator

try:
    import SherlockLifeCycleService_pb2
except ModuleNotFoundError:
    from ansys.api.sherlock.v0 import SherlockLifeCycleService_pb2


class HarmonicVibeProfileCsvFileProperties(BaseModel):
    """Properties of a harmonic vibe profile CSV file."""

    profile_name: str
    """Name of the harmonic vibe profile."""
    header_row_count: int
    """Number of rows before the column header in the file."""
    column_delimiter: str = ","
    """Delimiter used to separate columns in the file."""
    numeric_format: str = None
    """Numeric format for the values."""
    frequency_column: str
    """Name of the column containing frequency values."""
    frequency_units: str
    """Units of the frequency values"""
    load_column: str
    """Name of the column containing load values."""
    load_units: str
    """Units of the load values."""

    def _convert_to_grpc(
        self,
    ) -> SherlockLifeCycleService_pb2.LoadHarmonicProfileRequest.CSVProps:
        """Convert to gRPC CVSProps."""
        return SherlockLifeCycleService_pb2.LoadHarmonicProfileRequest.CSVProps(
            profileName=self.profile_name,
            headerRowNumber=self.header_row_count,
            columnDelim=self.column_delimiter,
            numericFormat=self.numeric_format,
            freqColumn=self.frequency_column,
            freqUnits=self.frequency_units,
            loadColumn=self.load_column,
            loadUnits=self.load_units,
        )

    @field_validator("header_row_count")
    @classmethod
    def non_negative_int_validation(cls, value: int, info: ValidationInfo):
        """Validate integer fields listed contain non-negative values."""
        if value < 0:
            raise ValueError(f"{info.field_name} must be greater than or equal to 0.")
        return value

    @field_validator(
        "profile_name", "frequency_column", "frequency_units", "load_column", "load_units"
    )
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    @field_validator("column_delimiter", "numeric_format")
    @classmethod
    def optional_str_validation(cls, value: Optional[str], info):
        """Allow the test_point_ids to not be set, i.e., None."""
        return optional_str_validator(value, info.field_name)


class RandomVibeProfileCsvFileProperties(BaseModel):
    """Properties of a random vibe profile CSV file."""

    profile_name: str
    """Name of the random vibe profile."""
    header_row_count: int
    """Number of rows before the column header in the file."""
    column_delimiter: str = ","
    """Delimiter used to separate columns in the file."""
    numeric_format: str = None
    """Numeric format for the values."""
    frequency_column: str
    """Name of the column containing frequency values."""
    frequency_units: str
    """Units of the frequency values"""
    amplitude_column: str
    """Name of the column containing amplitude values."""
    amplitude_units: str
    """Units of the amplitude values."""

    def _convert_to_grpc(
        self,
    ) -> SherlockLifeCycleService_pb2.LoadRandomVibeProfileRequest.CSVProps:
        """Convert to gRPC CVSProps."""
        return SherlockLifeCycleService_pb2.LoadRandomVibeProfileRequest.CSVProps(
            profileName=self.profile_name,
            headerRowNumber=self.header_row_count,
            columnDelim=self.column_delimiter,
            numericFormat=self.numeric_format,
            freqColumn=self.frequency_column,
            freqUnits=self.frequency_units,
            amplColumn=self.amplitude_column,
            amplUnits=self.amplitude_units,
        )

    @field_validator("header_row_count")
    @classmethod
    def non_negative_int_validation(cls, value: int, info: ValidationInfo):
        """Validate integer fields listed contain non-negative values."""
        if value < 0:
            raise ValueError(f"{info.field_name} must be greater than or equal to 0.")
        return value

    @field_validator(
        "profile_name", "frequency_column", "frequency_units", "amplitude_column", "amplitude_units"
    )
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    @field_validator("column_delimiter", "numeric_format")
    @classmethod
    def optional_str_validation(cls, value: Optional[str], info):
        """Allow the test_point_ids to not be set, i.e., None."""
        return optional_str_validator(value, info.field_name)


class ShockProfileDatasetCsvFileProperties(BaseModel):
    """Properties of a shock event profile using dataset CSV file."""

    profile_name: str
    """Name of the shock vibe profile."""
    header_row_count: int
    """Number of rows before the column header in the file."""
    column_delimiter: str = ","
    """Delimiter used to separate columns in the file."""
    numeric_format: str = None
    """Numeric format for the values."""
    time_column: str
    """Name of the column containing timeuency values."""
    time_units: str
    """Units of the timeuency values"""
    load_column: str
    """Name of the column containing load values."""
    load_units: str
    """Units of the load values."""

    def _convert_to_grpc(
        self,
    ) -> SherlockLifeCycleService_pb2.LoadShockProfileDatasetRequest.CSVProps:
        """Convert to gRPC CVSProps."""
        return SherlockLifeCycleService_pb2.LoadShockProfileDatasetRequest.CSVProps(
            profileName=self.profile_name,
            headerRowNumber=self.header_row_count,
            columnDelim=self.column_delimiter,
            numericFormat=self.numeric_format,
            timeColumn=self.time_column,
            timeUnits=self.time_units,
            loadColumn=self.load_column,
            loadUnits=self.load_units,
        )

    @field_validator("header_row_count")
    @classmethod
    def non_negative_int_validation(cls, value: int, info: ValidationInfo):
        """Validate integer fields listed contain non-negative values."""
        if value < 0:
            raise ValueError(f"{info.field_name} must be greater than or equal to 0.")
        return value

    @field_validator("profile_name", "time_column", "time_units", "load_column", "load_units")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    @field_validator("column_delimiter", "numeric_format")
    @classmethod
    def optional_str_validation(cls, value: Optional[str], info):
        """Allow the test_point_ids to not be set, i.e., None."""
        return optional_str_validator(value, info.field_name)


class ShockProfilePulsesCsvFileProperties(BaseModel):
    """Properties of a shock event profile using pulses CSV file."""

    profile_name: str
    """Name of the shock vibe profile."""
    header_row_count: int
    """Number of rows before the column header in the file."""
    column_delimiter: str = ","
    """Delimiter used to separate columns in the file."""
    numeric_format: str = None
    """Numeric format for the values."""
    duration: float
    """Pulse duration length"""
    duration_units: str
    """Time units of the pulse duration"""
    sample_rate: float
    """Sample rate"""
    sample_rate_units: str
    """Time units of the sample rate"""
    shape_column: str
    """Name of the column containing shape values."""
    load_column: str
    """Name of the column containing load values."""
    load_units: str
    """Units of the load values."""
    frequency_column: str
    """Name of the column containing frequency values."""
    frequency_units: str
    """Units of the frequency values"""
    decay_column: str
    """Name of the column containing decay values."""

    def _convert_to_grpc(
        self,
    ) -> SherlockLifeCycleService_pb2.LoadShockProfilePulsesRequest.CSVProps:
        """Convert to gRPC CVSProps."""
        return SherlockLifeCycleService_pb2.LoadShockProfilePulsesRequest.CSVProps(
            profileName=self.profile_name,
            headerRowNumber=self.header_row_count,
            columnDelim=self.column_delimiter,
            numericFormat=self.numeric_format,
            duration=self.duration,
            durationUnits=self.duration_units,
            sampleRate=self.sample_rate,
            sampleRateUnits=self.sample_rate_units,
            shapeColumn=self.shape_column,
            loadColumn=self.load_column,
            loadUnits=self.load_units,
            freqColumn=self.frequency_column,
            freqUnits=self.frequency_units,
            decayColumn=self.decay_column,
        )

    @field_validator("header_row_count")
    @classmethod
    def non_negative_int_validation(cls, value: int, info: ValidationInfo):
        """Validate integer fields listed contain non-negative values."""
        if value < 0:
            raise ValueError(f"{info.field_name} must be greater than or equal to 0.")
        return value

    @field_validator("duration", "sample_rate")
    @classmethod
    def greater_than_zero_float_validation(cls, value: float, info: ValidationInfo):
        """Validate float fields listed contain values greater than 0."""
        if value < 0:
            raise ValueError(f"{info.field_name} must be greater than 0.")
        return value

    @field_validator(
        "profile_name",
        "duration_units",
        "sample_rate_units",
        "shape_column",
        "load_column",
        "load_units",
        "frequency_column",
        "frequency_units",
        "decay_column",
    )
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    @field_validator("column_delimiter", "numeric_format")
    @classmethod
    def optional_str_validation(cls, value: Optional[str], info):
        """Allow the test_point_ids to not be set, i.e., None."""
        return optional_str_validator(value, info.field_name)


class ThermalProfileCsvFileProperties(BaseModel):
    """Properties of a thermal profile CSV file."""

    profile_name: str
    """Name of the thermal profile."""
    header_row_count: int
    """Number of rows before the column header in the file."""
    column_delimiter: str = ","
    """Delimiter used to separate columns in the file."""
    numeric_format: str = None
    """Numeric format for the values."""
    step_column: str
    """Name of the column containing step values."""
    type_column: str
    """Name of the column containing step type values."""
    time_column: str
    """Name of the column containing time duration values."""
    time_units: str
    """Units of the time values."""
    temperature_column: str
    """Name of the column containing temperature values."""
    temperature_units: str
    """Units of the temperature values."""

    def _convert_to_grpc(
        self,
    ) -> SherlockLifeCycleService_pb2.LoadThermalProfileRequest.CSVProps:
        """Convert to gRPC CVSProps."""
        return SherlockLifeCycleService_pb2.LoadThermalProfileRequest.CSVProps(
            profileName=self.profile_name,
            headerRowNumber=self.header_row_count,
            columnDelim=self.column_delimiter,
            numericFormat=self.numeric_format,
            stepColumn=self.step_column,
            typeColumn=self.type_column,
            timeColumn=self.time_column,
            timeUnits=self.time_units,
            tempColumn=self.temperature_column,
            tempUnits=self.temperature_units,
        )

    @field_validator("header_row_count")
    @classmethod
    def non_negative_int_validation(cls, value: int, info: ValidationInfo):
        """Validate integer fields listed contain non-negative values."""
        if value < 0:
            raise ValueError(f"{info.field_name} must be greater than or equal to 0.")
        return value

    @field_validator(
        "profile_name",
        "step_column",
        "type_column",
        "time_column",
        "time_units",
        "temperature_column",
        "temperature_units",
    )
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    @field_validator("column_delimiter", "numeric_format")
    @classmethod
    def optional_str_validation(cls, value: Optional[str], info):
        """Allow the test_point_ids to not be set, i.e., None."""
        return optional_str_validator(value, info.field_name)


class ThermalSignalFileProperties(BaseModel):
    """Properties of a thermal signal file."""

    header_row_count: int
    """Number of rows before the column header in the file."""
    numeric_format: str
    """Numeric format for the values."""
    column_delimiter: str
    """Delimiter used to separate columns in the file."""
    time_column: str
    """Name of the column containing time values."""
    time_units: str
    """Units of the time values."""
    temperature_column: str
    """Name of the column containing temperature values."""
    temperature_units: str
    """Units of the temperature values."""

    def _convert_to_grpc(
        self,
    ) -> SherlockLifeCycleService_pb2.ImportThermalSignalRequest.ThermalSignalFileProperties:
        """Convert to gRPC ThermalSignalFileProperties."""
        return SherlockLifeCycleService_pb2.ImportThermalSignalRequest.ThermalSignalFileProperties(
            headerRowCount=self.header_row_count,
            numericFormat=self.numeric_format,
            columnDelimiter=self.column_delimiter,
            timeColumn=self.time_column,
            timeUnits=self.time_units,
            temperatureColumn=self.temperature_column,
            temperatureUnits=self.temperature_units,
        )

    @field_validator("header_row_count")
    @classmethod
    def non_negative_int_validation(cls, value: int, info: ValidationInfo):
        """Validate integer fields listed contain non-negative values."""
        if value < 0:
            raise ValueError(f"{info.field_name} must be greater than or equal to 0.")
        return value

    @field_validator("time_column", "time_units", "temperature_column", "temperature_units")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)


class ImportThermalSignalRequest(BaseModel):
    """Request to import a thermal signal file."""

    file_name: str
    """The full path to the CSV thermal signal file to be imported."""
    project: str
    """Sherlock project name in which the thermal signal is imported."""
    phase_name: str
    """Name of the phase in which the thermal signal is imported."""
    thermal_signal_file_properties: ThermalSignalFileProperties
    """Properties of the thermal signal file."""
    time_removal: bool
    """Option to indicate that time results with shorter half-cycle durations are removed."""
    load_range_percentage: float
    """Defines the fraction of the range near peaks and valleys considered as a dwell region."""
    number_of_range_bins: int
    """Number of range bins for binning cycles, 0 for no range binning."""
    number_of_mean_bins: int
    """Number of mean bins for binning cycles, 0 for no mean binning."""
    number_of_dwell_bins: int
    """Number of dwell bins for binning cycles, 0 for no dwell binning."""
    temperature_range_filtering_limit: float
    """Minimum cycle range to include in results, 0 for not filtering."""
    time_filtering_limit: float
    """Maximum cycle time to include in results, default is 72 hours."""
    time_filtering_limit_units: str
    """Units of the time filtering limit."""
    generated_cycles_label: str
    """Label used to define the name of all generated thermal events."""

    @field_validator(
        "file_name", "project", "phase_name", "time_filtering_limit_units", "generated_cycles_label"
    )
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    @field_validator("number_of_range_bins", "number_of_mean_bins", "number_of_dwell_bins")
    @classmethod
    def non_negative_int_validation(cls, value: int, info: ValidationInfo):
        """Validate integer fields listed contain non-negative values."""
        if value < 0:
            raise ValueError(f"{info.field_name} must be greater than or equal to 0.")
        return value

    def _convert_to_grpc(self) -> SherlockLifeCycleService_pb2.ImportThermalSignalRequest:
        """Convert to gRPC ImportThermalSignalRequest."""
        return SherlockLifeCycleService_pb2.ImportThermalSignalRequest(
            thermalSignalFile=self.file_name,
            project=self.project,
            phaseName=self.phase_name,
            fileProperties=self.thermal_signal_file_properties._convert_to_grpc(),
            timeRemoval=self.time_removal,
            loadRangePercentage=self.load_range_percentage,
            numberOfRangeBins=self.number_of_range_bins,
            numberOfMeanBins=self.number_of_mean_bins,
            numberOfDwellBins=self.number_of_dwell_bins,
            temperatureRangeFilteringLimit=self.temperature_range_filtering_limit,
            timeFilteringLimit=self.time_filtering_limit,
            timeFilteringLimitUnits=self.time_filtering_limit_units,
            generatedCyclesLabel=self.generated_cycles_label,
        )


class UpdateLifeCycleRequest(BaseModel):
    """Request to update the life cycle."""

    project: str
    """Sherlock project name."""

    new_name: str
    """The new name of the life cycle."""

    new_description: str
    """The new description of the life cycle."""

    new_reliability_metric: float
    """The new reliability metric value.
    Options are: "Reliability (%)", "Prob. of Failure (%)",
    "MTBF (years)", "MTBF (hours)", "FITs (1E6 hrs)", "FITs (1E9 hrs)"
    """

    new_reliability_metric_units: str
    """The new reliability metric units."""

    new_service_life: float
    """The new service life value."""

    new_service_life_units: str
    """The new service life units.
    Options are: "year", "day", "hr", "min", "sec"
    """

    result_archive_file_name: str
    """File name for saved results. File names will be overwritten.
       Sub-Assembly results will be saved."""

    @field_validator("project")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> SherlockLifeCycleService_pb2.UpdateLifeCycleRequest:
        """Convert to gRPC SaveHarmonicProfileRequest."""
        return SherlockLifeCycleService_pb2.UpdateLifeCycleRequest(
            project=self.project,
            newName=self.new_name,
            newDescription=self.new_description,
            newReliabilityMetric=self.new_reliability_metric,
            newReliabilityMetricUnits=self.new_reliability_metric_units,
            newServiceLife=self.new_service_life,
            newServiceLifeUnits=self.new_service_life_units,
            resultArchiveFileName=self.result_archive_file_name,
        )


class SaveHarmonicProfileRequest(BaseModel):
    """Request to save a harmonic life cycle event profile to a .dat or .csv file."""

    project: str
    """Sherlock project name."""

    phase_name: str
    """The name of the life cycle phase this event is associated with."""

    event_name: str
    """Harmonic event name."""

    triaxial_axis: str | None = None
    """If the harmonic profile type is 'Triaxial', the axis this profile should be assigned to.
    Valid values are: x, y, z.
    """

    file_path: str
    """Full destination path for the .dat or .csv file."""

    @field_validator("project", "phase_name", "event_name", "file_path", "triaxial_axis")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> SherlockLifeCycleService_pb2.SaveHarmonicProfileRequest:
        """Convert to gRPC SaveHarmonicProfileRequest."""
        return SherlockLifeCycleService_pb2.SaveHarmonicProfileRequest(
            project=self.project,
            phaseName=self.phase_name,
            eventName=self.event_name,
            triaxialAxis=self.triaxial_axis or "",
            filePath=self.file_path,
        )


class SaveRandomVibeProfileRequest(BaseModel):
    """Request to save a random vibe life cycle event profile to a .dat or .csv file."""

    project: str
    """Sherlock project name."""

    phase_name: str
    """The name of the life cycle phase this event is associated with."""

    event_name: str
    """Random vibe event name."""

    file_path: str
    """Full destination path for the .dat or .csv file."""

    @field_validator("project", "phase_name", "event_name", "file_path")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> SherlockLifeCycleService_pb2.SaveRandomVibeProfileRequest:
        """Convert to gRPC SaveRandomVibeProfileRequest."""
        return SherlockLifeCycleService_pb2.SaveRandomVibeProfileRequest(
            project=self.project,
            phaseName=self.phase_name,
            eventName=self.event_name,
            filePath=self.file_path,
        )


class SaveShockPulseProfileRequest(BaseModel):
    """Request to save a shock pulse life cycle event profile to a .dat or .csv file."""

    project: str
    """Sherlock project name."""

    phase_name: str
    """The name of the life cycle phase this event is associated with."""

    event_name: str
    """Shock event name."""

    file_path: str
    """Full destination path for the .dat or .csv file."""

    @field_validator("project", "phase_name", "event_name", "file_path")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> SherlockLifeCycleService_pb2.SaveShockPulseProfileRequest:
        """Convert to gRPC SaveShockPulseProfileRequest."""
        return SherlockLifeCycleService_pb2.SaveShockPulseProfileRequest(
            project=self.project,
            phaseName=self.phase_name,
            eventName=self.event_name,
            filePath=self.file_path,
        )


class SaveThermalProfileRequest(BaseModel):
    """Request to save a thermal life cycle event profile to a .dat or .csv file."""

    project: str
    """Sherlock project name."""

    phase_name: str
    """The name of the life cycle phase this event is associated with."""

    event_name: str
    """Thermal event name."""

    file_path: str
    """Full destination path for the .dat or .csv file."""

    @field_validator("project", "phase_name", "event_name", "file_path")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> SherlockLifeCycleService_pb2.SaveThermalProfileRequest:
        """Convert to gRPC SaveThermalProfileRequest."""
        return SherlockLifeCycleService_pb2.SaveThermalProfileRequest(
            project=self.project,
            phaseName=self.phase_name,
            eventName=self.event_name,
            filePath=self.file_path,
        )


class DeleteEventRequest(BaseModel):
    """Request to delete a life cycle event from a project phase."""

    project: str
    """Sherlock project name."""

    phase_name: str
    """The name of the life cycle phase from which to delete this event."""

    event_name: str
    """Name of the event to be deleted."""

    @field_validator("project", "phase_name", "event_name")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> SherlockLifeCycleService_pb2.DeleteEventRequest:
        """Convert to gRPC DeleteEventRequest."""
        return SherlockLifeCycleService_pb2.DeleteEventRequest(
            project=self.project,
            phaseName=self.phase_name,
            eventName=self.event_name,
        )


class DeletePhaseRequest(BaseModel):
    """Request to delete a life cycle phase from a project."""

    project: str
    """Sherlock project name."""

    phase_name: str
    """Name of the life cycle phase to be deleted."""

    @field_validator("project", "phase_name")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> SherlockLifeCycleService_pb2.DeletePhaseRequest:
        """Convert to gRPC DeletePhaseRequest."""
        return SherlockLifeCycleService_pb2.DeletePhaseRequest(
            project=self.project,
            phaseName=self.phase_name,
        )


class UpdateLifePhaseRequest(BaseModel):
    """Request for updating an existing life phase in a specified project's life cycle."""

    project: str
    """Sherlock project name."""

    phase_name: str
    """Name of the life cycle phase to be updated."""

    new_phase_name: Optional[str] = None
    """(Optional) Updated name of life phase."""

    new_description: Optional[str] = None
    """(Optional) Updated description of life phase."""

    new_duration: Optional[float] = None
    """(Optional) Updated event duration length."""

    new_duration_units: Optional[str] = None
    """(Optional) Updated event duration units."""

    new_num_of_cycles: Optional[float] = None
    """(Optional) Updated number of cycles defined for the life phase."""

    new_cycle_type: Optional[str] = None
    """(Optional) Updated cycle type. Acceptable values are COUNT, DUTY_CYCLE, PER YEAR, PER DAY,
        PER HOUR, PER MIN, and PER SEC.
    """

    result_archive_file_name: Optional[str] = None
    """(Optional) Filename for saving analysis results and life cycle data, including any
        sub-assembly results. Any existing results will be overwritten. If this file name is
        omitted, then analysis results will not be saved during the update.
    """

    @field_validator("project", "phase_name")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    @field_validator(
        "new_phase_name",
        "new_description",
        "new_duration_units",
        "new_cycle_type",
        "result_archive_file_name",
    )
    @classmethod
    def optional_str_validation(cls, value: Optional[str], info):
        """Allow the optional fields to not be set, i.e., None."""
        return optional_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> SherlockLifeCycleService_pb2.UpdateLifePhaseRequest:
        request = SherlockLifeCycleService_pb2.UpdateLifePhaseRequest()
        request.project = self.project
        request.phaseName = self.phase_name
        if self.new_phase_name is not None:
            request.newPhaseName = self.new_phase_name
        if self.new_description is not None:
            request.newDescription = self.new_description
        if self.new_duration is not None:
            request.newDuration = self.new_duration
        if self.new_duration_units is not None:
            request.newDurationUnits = self.new_duration_units
        if self.new_num_of_cycles is not None:
            request.newNumOfCycles = self.new_num_of_cycles
        if self.new_cycle_type is not None:
            request.newCycleType = self.new_cycle_type
        if self.result_archive_file_name is not None:
            request.resultArchiveFileName = self.result_archive_file_name
        return request
